/*
* Copyright (c) 2008-2017, NVIDIA CORPORATION. All rights reserved.
*
* NVIDIA CORPORATION and its licensors retain all intellectual property
* and proprietary rights in and to this software, related documentation
* and any modifications thereto. Any use, reproduction, disclosure or
* distribution of this software and related documentation without an express
* license agreement from NVIDIA CORPORATION is strictly prohibited.
*/

#include "./test_util.h"
#include <nvrhi/nvrhi.h>

#include <filesystem>

std::filesystem::path FindDirectoryWithFile(const std::filesystem::path& startPath, const std::filesystem::path& relativeFilePath, int maxDepth)
{
	std::filesystem::path searchPath = "";

	for (int depth = 0; depth < maxDepth; depth++)
	{
		std::filesystem::path currentPath = startPath / searchPath / relativeFilePath;

		if (std::filesystem::is_directory(currentPath))
		{
			return currentPath.parent_path().lexically_normal();
		}

		searchPath = ".." / searchPath;
	}

	return "";
}

static std::filesystem::path g_mediaFolder = FindDirectoryWithFile(std::filesystem::current_path(), "media", 5) / "media";

std::filesystem::path const& GetMediaFolder()
{
	return g_mediaFolder;
}

GenericImageType* LoadTexture(char const* Filename)
{
	GenericImageType *genericImage;
	DDSURFACEDESC2 ddsd;
	char filecode[4];
	FILE *fp;
	int factor, uncompressed_components, format_found;

	// try to open the file 
#ifdef WIN32
	fopen_s(&fp, Filename, "rb");
#else
	fp = fopen(Filename, "rb");
#endif
	if (fp == nullptr)
	{
		fprintf(stderr, "\nUnable to load texture file [%s]", Filename);
		return nullptr;
	}

	// verify the type of file 
	fread(filecode, 1, 4, fp);
	if (strncmp(filecode, "DDS ", 4) != 0)
	{
		fprintf(stderr, "\nThe file [%s] is not a dds file", Filename);
		fclose(fp);
		return nullptr;
	}

	// get the surface desc 
	fread(&ddsd, sizeof(ddsd), 1, fp);
	if ((ddsd.dwFlags & DDSD_PIXELFORMAT) != DDSD_PIXELFORMAT)
	{
		fprintf(stderr, "\nDDS header pixelformat field contains no valid data,\nunable to load texture (%s).", Filename);
		return nullptr;
	}
	if (((ddsd.dwFlags & DDSD_WIDTH) != DDSD_WIDTH) || ((ddsd.dwFlags & DDSD_HEIGHT) != DDSD_HEIGHT))
	{
		fprintf(stderr, "\nDDS header width/height fields contains no valid data,\nunable to load texture (%s).", Filename);
		return nullptr;
	}
	if ((ddsd.dwFlags & DDSD_MIPMAPCOUNT) != DDSD_MIPMAPCOUNT)
	{
		fprintf(stderr, "\nDDS header mipmapcount field contains no valid data,\nassuming 0 mipmaps (%s).", Filename);
		//return 0;
	}
	if (((ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC) != DDPF_FOURCC) && ((ddsd.ddpfPixelFormat.dwFlags & DDPF_RGB) != DDPF_RGB))
	{
		fprintf(stderr, "\nDDS header pixelformat field contains no valid FOURCC and RGB data,\nunable to load texture (%s).", Filename);
		return nullptr;
	}
	if (((ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC) == DDPF_FOURCC) && (ddsd.dwMipMapCount <= 1) && ((ddsd.ddpfPixelFormat.dwFourCC == FOURCC_DXT1) || (ddsd.ddpfPixelFormat.dwFourCC == FOURCC_DXT3) || (ddsd.ddpfPixelFormat.dwFourCC == FOURCC_DXT5)))
	{
		fprintf(stderr, "\nDDS header contains DXTx FOURCC code and no mipmaps,\nprogram does not support loading DXTx textures without mipmaps,\nunable to load texture (%s).", Filename);
		return nullptr;
	}
	genericImage = (GenericImageType*)malloc(sizeof(GenericImageType));
	memset(genericImage, 0, sizeof(GenericImageType));
	factor = 0;
	uncompressed_components = 0;
	format_found = 0;
	genericImage->width = ddsd.dwWidth;
	genericImage->height = ddsd.dwHeight;
	if ((ddsd.ddpfPixelFormat.dwFlags & DDPF_FOURCC) == DDPF_FOURCC) // if FOURCC code is valid
	{
		genericImage->format = ddsd.ddpfPixelFormat.dwFourCC;
		switch (ddsd.ddpfPixelFormat.dwFourCC)
		{
		case FOURCC_DXT1:
			if (ddsd.dwAlphaBitDepth != 0)
			{
				genericImage->components = 4;
			}
			else
			{
				genericImage->components = 3;
			}
			factor = 2;
			format_found = 1;
			break;
		case FOURCC_DXT3:
			genericImage->components = 4;
			factor = 4;
			format_found = 1;
			break;
		case FOURCC_DXT5:
			genericImage->components = 4;
			factor = 4;
			format_found = 1;
			break;
		default:
			break;
		}
	}

	if (format_found == 0) // Uncompressed format apparently
	{
		switch (ddsd.ddpfPixelFormat.dwRGBBitCount)
		{
		case 24:
			genericImage->format = 0;
			genericImage->components = 3;
			format_found = 2;
			break;
		case 32:
			genericImage->format = 0;
			genericImage->components = 4;
			format_found = 2;
		default:
			break;
		}
	}
	if (format_found == 1)  // found compressed format
	{
		genericImage->dataSize = ddsd.dwMipMapCount > 1 ? ddsd.dwLinearSize * factor  + (ddsd.dwLinearSize * factor / 2) : ddsd.dwLinearSize * factor;
		genericImage->baseRowPitch = ddsd.dwWidth * factor;
	}
	if (format_found == 2)
	{
		genericImage->dataSize = ddsd.dwMipMapCount > 1 ? ddsd.dwWidth*ddsd.dwHeight*genericImage->components + (ddsd.dwWidth*ddsd.dwHeight*genericImage->components) / 2 : ddsd.dwWidth*ddsd.dwHeight*genericImage->components;
		genericImage->baseRowPitch = ddsd.dwWidth*genericImage->components;
	}
	if (format_found == 0)
	{
		fprintf(stderr, "\nUnsupported DDS format,\nthe program only supports DXTx and RGB/RGBA formats containing 8 bits per component");
		free(genericImage);
		return nullptr;
	}

	genericImage->pixels = (unsigned char*)malloc(genericImage->dataSize);
	fread(genericImage->pixels, 1, genericImage->dataSize, fp);
	genericImage->numMipmaps = ddsd.dwMipMapCount > 0 ? ddsd.dwMipMapCount : 1;

	return genericImage;
}

nvrhi::TextureHandle CreateTextureFromFile(nvrhi::IDevice* pDevice, nvrhi::ICommandList* pCommandList, char const* filename)
{
	GenericImageType* genericImage;
	nvrhi::TextureHandle result;

	genericImage = LoadTexture(filename);
	if (genericImage == nullptr) return nullptr;

	nvrhi::TextureDesc desc;
	desc.width = genericImage->width;
	desc.height = genericImage->height;
	desc.mipLevels = genericImage->numMipmaps;
	desc.format = nvrhi::Format::RGBA8_UNORM;
	desc.debugName = filename;

	if(genericImage->components != 4)
	{
		fprintf(stderr, "\nLoading textures with 3-component (RGB) formats is not supported");
		free(genericImage->pixels);
		free(genericImage);
		return nullptr;
	}
	if (genericImage->format != 0)
	{
		fprintf(stderr, "\nLoading textures with compressed formats is not supported");
		free(genericImage->pixels);
		free(genericImage);
		return nullptr;
	}

	result = pDevice->createTexture(desc);
	pCommandList->beginTrackingTextureState(result, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);

	uint8_t *data = genericImage->pixels;

	for (uint32_t i = 0; i < desc.mipLevels; i++)
	{
		pCommandList->writeTexture(result, 0, i, data, genericImage->baseRowPitch);
		
		data += desc.height * genericImage->baseRowPitch;

		genericImage->baseRowPitch >>= 1;
		desc.width >>= 1;
		desc.height >>= 1;
	} 

	free(genericImage->pixels);
	free(genericImage);
	pCommandList->endTrackingTextureState(result, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE, true);

	return result;
}
