/*
* Copyright (c) 2008-2017, NVIDIA CORPORATION. All rights reserved.
*
* NVIDIA CORPORATION and its licensors retain all intellectual property
* and proprietary rights in and to this software, related documentation
* and any modifications thereto. Any use, reproduction, disclosure or
* distribution of this software and related documentation without an express
* license agreement from NVIDIA CORPORATION is strictly prohibited.
*/
#include "./ui_renderer.h"
#include "./test_util.h"

#if USE_VK
	#include <shaders/bytecode/ImGUIVS_vk.h>
	#include <shaders/bytecode/ImGUIPS_vk.h>
#else
	#include <shaders/bytecode/ImGUIVS_dx.h>
	#include <shaders/bytecode/ImGUIPS_dx.h>
#endif // USE_VK


struct IMGUI_VS_CBUFFER
{
	float        mvp[4][4];
};

uint32_t IndexToGridSize(uint32_t DetailLevelIndex)
{
	switch (DetailLevelIndex)
	{
		case 0: return 128;
		case 1: return 256;
		case 2: return 512;
		case 3: return 1024;
		case 4: return 2048;
		default: return 128;
	}
}

uint32_t GridSizeToIndex(uint32_t GridSize)
{
	switch (GridSize)
	{
		case 128: return 0;
		case 256: return 1;
		case 512: return 2;
		case 1024: return 3;
		case 2048: return 4;
		default: return 0;
	}
}

void UIRenderer::BuildUI(void)
{
    const auto& io = ImGui::GetIO();
	float ColumnOffset = 250.0f;
	ImGui::SetNextWindowPos(ImVec2(10, 10), ImGuiSetCond_Always);
	ImGui::SetNextWindowSizeConstraints(ImVec2(0,0), ImVec2(500, io.DisplaySize.y - 20));

	ImGui::Begin("WaveWorks ", 0, ImGuiWindowFlags_AlwaysAutoResize);

	if (ImGui::TreeNode("General settings"))
	{

		ImGui::Checkbox("Simulate water", &(m_pState->bSimulateWater));
		ImGui::Checkbox("Render water", &(m_pState->bRenderWater));
		ImGui::Checkbox("Use wireframe", &(m_pState->bRenderWireframe));
		ImGui::Checkbox("Use microfacet Fresnel", &(m_pState->bUseMicrofacetFresnel));
		ImGui::Checkbox("Use microfacet Specular", &(m_pState->bUseMicrofacetSpecular));
		ImGui::Checkbox("Use microfacet Reflection", &(m_pState->bUseMicrofacetReflection));
		ImGui::NewLine();
		if(ImGui::SliderFloat("UV warping amplitude", &(m_pState->OceanWindSimulationParameters.uv_warping_amplitude), 0.0f, 0.1f, "%.2f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
		if(ImGui::SliderFloat("UV warping frequency", &(m_pState->OceanWindSimulationParameters.uv_warping_frequency), 1.0f, 3.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
		ImGui::Checkbox("Show cascades", &(m_pState->bShowCascades));
		ImGui::NewLine();
		if (ImGui::Checkbox("Use atmosphere model", &(m_pState->bUseDynamicSky)))
		{
			m_pState->bNeedToUpdateSky = true;
		}
		
		if (m_pState->bUseDynamicSky)
		{
			if (ImGui::SliderFloat("Sun position", &(m_pState->fSunAngle), 1.0f, 90.0f, "%.0f", 1.0f))
			{
				m_pState->bNeedToUpdateSky = true;
			}
		}
		else
		{
			m_pState->fSunAngle = 45.0;
		}

		ImGui::SliderFloat("Sun intensity", &(m_pState->fSunIntensity), 0.0f, 2.0f, "%.2f", 1.0f);
		ImGui::SliderFloat("Beckmann roughness", &(m_pState->fBeckmannRoughness), 0.00001f, 0.1f, "%.5f", 2.0f);
		ImGui::NewLine();
		ImGui::TreePop();
	}
	if (ImGui::TreeNode("Wind waves"))
	{
		if (ImGui::TreeNode("Simulation settings"))
		{
			if (ImGui::SliderFloat("Simulation period (m)", &(m_pState->OceanWindSimulationSettings.simulation_period), 100.0f, 10000.0f, "%.0f", 2.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			if (ImGui::Checkbox("Use Beaufort scale", &(m_pState->OceanWindSimulationSettings.use_Beaufort_scale))) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;

			if (ImGui::Checkbox("Enable CPU driven displacement readbacks", &(m_pState->OceanWindSimulationSettings.enable_CPU_driven_displacement_calculation))) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;

			if (ImGui::Checkbox("Enable GPU driven displacement readbacks", &(m_pState->OceanWindSimulationSettings.enable_GPU_driven_displacement_calculation))) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;

			char*  ReadbackUsageStrings[3] = { "None", "CPU", "Compute" };
			ImGui::Combo("Use displacement readbacks", &m_pState->iReadbackUsage, ReadbackUsageStrings, 3);


			if (ImGui::Checkbox("Use CPU timers", &(m_pState->OceanWindSimulationSettings.enable_CPU_timers)))
			{
				m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}

			if (ImGui::Checkbox("Use GPU timers", &(m_pState->OceanWindSimulationSettings.enable_GPU_timers)))
			{
				m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}

			if (m_pState->iReadbackUsage > 0)
			{
				ImGui::Checkbox("Render readback markers", &(m_pState->bRenderMarkers));
				ImGui::SliderInt("Refinement steps", &(m_pState->iRefinementSteps), 1, 20, "%.0f");
				ImGui::SliderFloat("ReadbackCoord", &(m_pState->fReadbackCoord), 0, 30, "%.1f");
			}
			else
			{
				m_pState->bRenderMarkers = false;
			}

			char*  SimulationAPIStrings[2] = { "CPU", "Compute" };
			int SimulationAPIIndex = (int)(m_pState->OceanWindSimulationSettings.simulation_api);
			if (ImGui::Combo("Simulation API", &SimulationAPIIndex, SimulationAPIStrings, 2))
			{
				m_pState->OceanWindSimulationSettings.simulation_api = (GFSDK_WaveWorks_Simulation_API)SimulationAPIIndex;
				m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}

			char*  DetailLevelStrings[3] = { "Normal", "High", "Extreme" };
			int DetailLevelIndex = (int)(m_pState->OceanWindSimulationSettings.detail_level);
			if (ImGui::Combo("Simulation Quality", &DetailLevelIndex, DetailLevelStrings, 3))
			{
				m_pState->OceanWindSimulationSettings.detail_level = (GFSDK_WaveWorks_Simulation_DetailLevel)DetailLevelIndex;
				m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}

			ImGui::NewLine();
			ImGui::TreePop();

		}

		if (ImGui::TreeNode("Simulation parameters"))
		{
			if (ImGui::TreeNode("Local wind"))
			{
				if (ImGui::SliderFloat("Wind direction (degrees)", &(m_pState->fBaseWindDirection), -180.0f, 180.0f, "%.0f", 1.0f))
				{
					m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
					m_pState->OceanWindSimulationParameters.base_wind_direction.x = cosf(m_pState->fBaseWindDirection * 0.0174532925f); // degrees to radians
					m_pState->OceanWindSimulationParameters.base_wind_direction.y = sinf(m_pState->fBaseWindDirection * 0.0174532925f); // degrees to radians
				}

				if (m_pState->OceanWindSimulationSettings.use_Beaufort_scale)
				{
					if (m_pState->OceanWindSimulationParameters.base_wind_speed > 12.0f) m_pState->OceanWindSimulationParameters.base_wind_speed = 12.0f;
					if (ImGui::SliderFloat("Wind speed (Beaufort)", &(m_pState->OceanWindSimulationParameters.base_wind_speed), 0, 12.0f, "%.1f", 1.0f))
					{
						m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
						m_pState->fMaxTotalDisplacement = 0;
						m_pState->fMinTotalDisplacement = 0;
					}
				}
				else
				{
					if (ImGui::SliderFloat("Wind speed (m/sec)", &(m_pState->OceanWindSimulationParameters.base_wind_speed), 0, 50.0f, "%.1f", 1.0f))
					{
						m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
						m_pState->fMaxTotalDisplacement = 0;
						m_pState->fMinTotalDisplacement = 0;
					}
					if (ImGui::SliderFloat("Wind distance (km)", &(m_pState->OceanWindSimulationParameters.base_wind_distance), 1.0f, 1000.0f, "%.0f", 2.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
					if (ImGui::SliderFloat("Spectrum peaking", &(m_pState->OceanWindSimulationParameters.base_spectrum_peaking), 0.1f, 10.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
					if (ImGui::SliderFloat("Amplitude", &(m_pState->OceanWindSimulationParameters.base_amplitude_multiplier), 0, 2.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				}
				if (ImGui::SliderFloat("Wind dependency", &(m_pState->OceanWindSimulationParameters.base_wind_dependency), 0, 1.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Small waves lengths (m)", &(m_pState->OceanWindSimulationParameters.base_small_waves_cutoff_length), 0, 100.0f, "%.1f", 2.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Small waves cutoff", &(m_pState->OceanWindSimulationParameters.base_small_waves_cutoff_power), 0, 1.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				ImGui::TreePop();
			}

			if (ImGui::TreeNode("Swell"))
			{
				if (ImGui::SliderFloat("Wind direction (degrees)", &(m_pState->fSwellWindDirection), -180.0f, 180.0f, "%.0f", 1.0f))
				{
					m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
					m_pState->OceanWindSimulationParameters.swell_wind_direction.x = cosf(m_pState->fSwellWindDirection * 0.0174532925f); // degrees to radians
					m_pState->OceanWindSimulationParameters.swell_wind_direction.y = sinf(m_pState->fSwellWindDirection * 0.0174532925f); // degrees to radians
				}
				if (ImGui::SliderFloat("Wind speed (m/sec)", &(m_pState->OceanWindSimulationParameters.swell_wind_speed), 0, 50.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Wind distance (km)", &(m_pState->OceanWindSimulationParameters.swell_wind_distance), 1.0f, 1000.0f, "%.0f", 2.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Spectrum peaking", &(m_pState->OceanWindSimulationParameters.swell_spectrum_peaking), 0.1f, 10.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Amplitude", &(m_pState->OceanWindSimulationParameters.swell_amplitude_multiplier), 0, 2.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Wind dependency", &(m_pState->OceanWindSimulationParameters.swell_wind_dependency), 0, 1.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Small waves lengths (m)", &(m_pState->OceanWindSimulationParameters.swell_small_waves_cutoff_length), 0, 100.0f, "%.1f", 2.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Small waves cutoff", &(m_pState->OceanWindSimulationParameters.swell_small_waves_cutoff_power), 0, 1.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				ImGui::TreePop();
			}


			ImGui::Text("Common parameters:");
			if (!m_pState->OceanWindSimulationSettings.use_Beaufort_scale)
			{

				if (ImGui::SliderFloat("Choppiness", &(m_pState->OceanWindSimulationParameters.lateral_multiplier), 0, 4.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}
			if (ImGui::SliderFloat("Time scale", &(m_pState->OceanWindSimulationParameters.time_scale), 0, 5.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;

			if (!m_pState->OceanWindSimulationSettings.use_Beaufort_scale)
			{
				ImGui::Text("Foam generation parameters:");
				if (ImGui::SliderFloat("Whitecaps threshold", &(m_pState->OceanWindSimulationParameters.foam_whitecaps_threshold), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Generation threshold", &(m_pState->OceanWindSimulationParameters.foam_generation_threshold), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Generation amount", &(m_pState->OceanWindSimulationParameters.foam_generation_amount), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Dissipation speed", &(m_pState->OceanWindSimulationParameters.foam_dissipation_speed), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
				if (ImGui::SliderFloat("Falloff speed", &(m_pState->OceanWindSimulationParameters.foam_falloff_speed), 0.95f, 0.99f, "%.3f", 1.0f)) m_pState->bNeedToUpdateWindWavesSimulationProperties = true;
			}
			ImGui::NewLine();
			ImGui::TreePop();
		}

		ImGui::Checkbox("Show statistics", &(m_pState->bShowWindWavesSimulationStatistics));
		if (m_pState->bShowWindWavesSimulationStatistics)
		{
			ImGui::Text("CPU_main_thread_wait_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanWindSimulationStatsFiltered.CPU_main_thread_wait_time);
			ImGui::Text("CPU_threads_total_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanWindSimulationStatsFiltered.CPU_threads_total_time);
			ImGui::Text("GPU_simulation_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanWindSimulationStatsFiltered.GPU_simulation_time);
			ImGui::Text("GPU_update_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanWindSimulationStatsFiltered.GPU_update_time);
			ImGui::Text("GPU_total_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanWindSimulationStatsFiltered.GPU_total_time);
			ImGui::Text("Lask simulation KickID"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iLastWindSimulationKickID);
			ImGui::Text("Lask readback KickID"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iLastWindSimulationReadbackKickID);
			ImGui::Text("Readback latency"); ImGui::SameLine(ColumnOffset); m_pState->iWindSimulationReadbackLatency != -1 ? ImGui::Text("%i", m_pState->iWindSimulationReadbackLatency) : ImGui::Text("N/A");
			ImGui::Text("Rendering latency"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iWindSimulationRenderLatency);
			float fmaxdisp = m_pState->fMaxDisplacement - m_pState->fMinDisplacement;
			float fmaxtotaldisp = m_pState->fMaxTotalDisplacement - m_pState->fMinTotalDisplacement;
			ImGui::Text("Min/Max displ, max wave height"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.2f    %3.2f    %3.2f", m_pState->fMinDisplacement, m_pState->fMaxDisplacement, fmaxdisp); ImGui::NewLine();
			ImGui::Text("Max wave height total"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.2f", fmaxtotaldisp); ImGui::NewLine();
			ImGui::Text("Significant Wave Height"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.2f", m_pState->fSWH); ImGui::NewLine();
			float p = GFSDK_WaveWorks_Wind_Waves_Simulation_GetConservativeMaxDisplacementEstimate(m_pState->hOceanWindSimulation);
			ImGui::Text("Conservative displacement estimation"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.2f", GFSDK_WaveWorks_Wind_Waves_Simulation_GetConservativeMaxDisplacementEstimate(m_pState->hOceanWindSimulation)); ImGui::NewLine();
		}

		ImGui::NewLine();
		ImGui::TreePop();
	}

	// Local waves simulation
	if (ImGui::TreeNode("Local waves"))
	{
		if (ImGui::TreeNode("Simulation settings"))
		{
			if (ImGui::SliderFloat("Simulation domain size (m)", &(m_pState->OceanLocalSimulationSettings.simulation_domain_worldspace_size), 10.0f, 1000.0f, "%.0f", 2.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;

			if (ImGui::Checkbox("Enable CPU driven displacement readbacks", &(m_pState->OceanLocalSimulationSettings.enable_CPU_driven_displacement_calculation))) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;

			if (ImGui::Checkbox("Enable GPU driven displacement readbacks", &(m_pState->OceanLocalSimulationSettings.enable_GPU_driven_displacement_calculation))) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			
			char*  ReadbackUsageStrings[3] = { "None", "CPU", "Compute" };
			ImGui::Combo("Use displacement readbacks", &m_pState->iReadbackUsage, ReadbackUsageStrings, 3);

			if (ImGui::Checkbox("Use CPU timers", &(m_pState->OceanLocalSimulationSettings.enable_CPU_timers)))
			{
				m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			}

			if (ImGui::Checkbox("Use GPU timers", &(m_pState->OceanLocalSimulationSettings.enable_GPU_timers)))
			{
				m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			}

			char*  SimulationAPIStrings[2] = { "CPU", "Compute" };
			int SimulationAPIIndex = (int)(m_pState->OceanLocalSimulationSettings.simulation_api);
			if (ImGui::Combo("Simulation API", &SimulationAPIIndex, SimulationAPIStrings, 2))
			{
				m_pState->OceanLocalSimulationSettings.simulation_api = (GFSDK_WaveWorks_Simulation_API)SimulationAPIIndex;
				m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			}

			char*  DetailLevelStrings[5] = { "128", "256", "512", "1024", "2048" };
			int DetailLevelIndex = (int)(GridSizeToIndex(m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size));
			if (ImGui::Combo("Simulation grid size", &DetailLevelIndex, DetailLevelStrings, 5))
			{
				m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size = IndexToGridSize(DetailLevelIndex);
				m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			}
			if (ImGui::Button("Reset simulation")) m_pState->bNeedToResetLocalWavesSimulation = true;
			if (ImGui::Button("Add disturbances"))
			{
				gfsdk_float4 d;
				for(int i=-15; i< 10; i++)
					for (int j = -15; j < 10; j++)
					{
						d.x = (float)(i + m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size / 2);
						d.y = (float)(j + m_pState->OceanLocalSimulationSettings.simulation_domain_grid_size / 2);
						d.z = 1.0f;
						d.w = 0;
						GFSDK_WaveWorks_Local_Waves_Simulation_AddDisturbances(m_pState->hOceanLocalSimulation, &d, 1);
					}
			}
			ImGui::Checkbox("Rain effect", &(m_pState->bRain));
			ImGui::SliderFloat("Rain Drop Size", &(m_pState->fRainDropSize), 0.01f, 0.2f, "%.2f", 1.0f);
			ImGui::Checkbox("Simulate \"boat\" ", &(m_pState->bBoat));

			ImGui::NewLine();
			ImGui::TreePop();

		}

		if (ImGui::TreeNode("Simulation parameters"))
		{
			if (ImGui::SliderFloat("Amplitude", &(m_pState->OceanLocalSimulationParameters.amplitude_multiplier), 0, 2.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			if (ImGui::SliderFloat("Choppiness", &(m_pState->OceanLocalSimulationParameters.lateral_multiplier), 0, 2.0f, "%.1f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;

			ImGui::Text("Foam generation parameters:");
			if (ImGui::SliderFloat("Whitecaps threshold", &(m_pState->OceanLocalSimulationParameters.foam_whitecaps_threshold), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			if (ImGui::SliderFloat("Generation threshold", &(m_pState->OceanLocalSimulationParameters.foam_generation_threshold), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			if (ImGui::SliderFloat("Generation amount", &(m_pState->OceanLocalSimulationParameters.foam_generation_amount), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			if (ImGui::SliderFloat("Dissipation speed", &(m_pState->OceanLocalSimulationParameters.foam_dissipation_speed), 0.f, 1.0f, "%.2f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;
			if (ImGui::SliderFloat("Falloff speed", &(m_pState->OceanLocalSimulationParameters.foam_falloff_speed), 0.5f, 0.99f, "%.3f", 1.0f)) m_pState->bNeedToUpdateLocalWavesSimulationProperties = true;

			ImGui::NewLine();
			ImGui::TreePop();
		}
		ImGui::Checkbox("Show statistics", &(m_pState->bShowLocalWavesSimulationStatistics));
		if (m_pState->bShowLocalWavesSimulationStatistics)
		{
			ImGui::NewLine();
			ImGui::Text("CPU_main_thread_wait_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanLocalSimulationStatsFiltered.CPU_main_thread_wait_time);
			ImGui::Text("CPU_threads_total_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanLocalSimulationStatsFiltered.CPU_threads_total_time);
			ImGui::Text("GPU_simulation_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanLocalSimulationStatsFiltered.GPU_simulation_time);
			ImGui::Text("GPU_update_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanLocalSimulationStatsFiltered.GPU_update_time);
			ImGui::Text("GPU_total_time"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.6f msec", m_pState->OceanLocalSimulationStatsFiltered.GPU_total_time);
			ImGui::Text("Lask simulation KickID"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iLastLocalSimulationKickID);
			ImGui::Text("Lask readback KickID"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iLastLocalSimulationReadbackKickID);
			ImGui::Text("Readback latency"); ImGui::SameLine(ColumnOffset); m_pState->iLocalSimulationReadbackLatency != -1 ? ImGui::Text("%i", m_pState->iLocalSimulationReadbackLatency) : ImGui::Text("N/A");
			ImGui::Text("Rendering latency"); ImGui::SameLine(ColumnOffset); ImGui::Text("%i", m_pState->iLocalSimulationRenderLatency);
		}
		ImGui::NewLine();
		ImGui::TreePop();
	}

	if (ImGui::TreeNode("Quadtree / geometry parameters"))
	{
		int cellCount = (int)m_pState->OceanQuadtreeParameters.cell_count;
		if (ImGui::SliderInt("Cells in single patch", &cellCount, 8, 256, "%.0f"))
		{
			m_pState->OceanQuadtreeParameters.cell_count = (uint32_t)cellCount;
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}

		if (ImGui::SliderFloat("Minimum patch size (m)", &(m_pState->OceanQuadtreeParameters.min_patch_length), 1.0f, 100.0f, "%.0f", 1.0f))
		{
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}

		if (ImGui::SliderFloat("Maximum edge length (pixels)", &(m_pState->OceanQuadtreeParameters.max_edge_length), 5.0f, 50.0f, "%.0f", 1.0f))
		{
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}

		if (ImGui::SliderFloat("Mean sea level (m)", &(m_pState->OceanQuadtreeParameters.mean_sea_level), -50.0f, 50.0f, "%.1f", 1.0f))
		{
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}

		if (ImGui::SliderFloat("Geomorphing degree", &(m_pState->OceanQuadtreeParameters.geomorphing_degree), 0.0f, 1.0f, "%.2f", 1.0f))
		{
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}

		if (ImGui::Checkbox("Generate diamond pattern", &(m_pState->OceanQuadtreeParameters.generate_diamond_pattern)))
		{
			m_pState->bNeedToUpdateQuadtreeProperties = true;
		}
		
		ImGui::NewLine();
		ImGui::SliderFloat("Static tessellation offset", &(m_pState->fTessellationOffset), 0.0f, 16.0f, "%.0f", 1.0f);
		ImGui::SliderFloat("Dynamic tessellation amount", &(m_pState->fTessellationAmount), 0.1f, 100.0f, "%.2f", 1.0f);

		ImGui::NewLine();
		ImGui::TreePop();
	}
	
	ImGui::NewLine();
	ImGui::Text("Renderer: %s", m_pDeviceManager->GetRendererString());
	ImGui::Text("%.3f msec/frame (%.1f FPS)", 1000.f / io.Framerate, io.Framerate);
	ImGui::Text("Camera pos"); ImGui::SameLine(ColumnOffset); ImGui::Text("%3.1f    %3.1f    %3.1f", m_pState->Camera.GetCameraPosition().x, m_pState->Camera.GetCameraPosition().y, m_pState->Camera.GetCameraPosition().z);
	ImGui::Text("Quadtree stats:"); 
	ImGui::Text("Nodes: %i total, %i rendered, time : %3.3f msec", m_pState->OceanQuadtreeStats.num_quadtree_nodes_created, m_pState->OceanQuadtreeStats.num_quadtree_nodes_for_rendering, m_pState->OceanQuadtreeStats.CPU_quadtree_update_time);
	ImGui::Text("Refinement iterations: %i, nodes added: %i", m_pState->OceanQuadtreeStats.num_refinement_iterations, m_pState->OceanQuadtreeStats.num_refinement_nodes_added);

	ImGui::End();
}

bool UIRenderer::Init()
{
    // Set up keyboard mapping.
    // ImGui will use those indices to peek into the io.KeyDown[] array
    // that we will update during the application lifetime.
    ImGuiIO& io = ImGui::GetIO();
    io.KeyMap[ImGuiKey_Tab] = GLFW_KEY_TAB;
    io.KeyMap[ImGuiKey_LeftArrow] = GLFW_KEY_LEFT;
    io.KeyMap[ImGuiKey_RightArrow] = GLFW_KEY_RIGHT;
    io.KeyMap[ImGuiKey_UpArrow] = GLFW_KEY_UP;
    io.KeyMap[ImGuiKey_DownArrow] = GLFW_KEY_DOWN;
    io.KeyMap[ImGuiKey_PageUp] = GLFW_KEY_PAGE_UP;
    io.KeyMap[ImGuiKey_PageDown] = GLFW_KEY_PAGE_DOWN;
    io.KeyMap[ImGuiKey_Home] = GLFW_KEY_HOME;
    io.KeyMap[ImGuiKey_End] = GLFW_KEY_END;
    io.KeyMap[ImGuiKey_Delete] = GLFW_KEY_DELETE;
    io.KeyMap[ImGuiKey_Backspace] = GLFW_KEY_BACKSPACE;
    io.KeyMap[ImGuiKey_Enter] = GLFW_KEY_ENTER;
    io.KeyMap[ImGuiKey_Escape] = GLFW_KEY_ESCAPE;
    io.KeyMap[ImGuiKey_A] = 'A';
    io.KeyMap[ImGuiKey_C] = 'C';
    io.KeyMap[ImGuiKey_V] = 'V';
    io.KeyMap[ImGuiKey_X] = 'X';
    io.KeyMap[ImGuiKey_Y] = 'Y';
    io.KeyMap[ImGuiKey_Z] = 'Z';

	io.Fonts->AddFontFromFileTTF((GetMediaFolder() / "DroidSans.ttf").generic_string().c_str(), 15.f, nullptr, nullptr);

	ImGui::PushStyleVar(ImGuiStyleVar_WindowRounding, 0.0f);
	ImGui::PushStyleVar(ImGuiStyleVar_ItemSpacing, ImVec2(2.0f, 2.0f));


	nvrhi::CommandListParameters cmdListParams;
	m_hCommandList = m_pDevice->createCommandList(cmdListParams);
	m_hCommandList->open();

	// Loading shaders
#if USE_VK
	auto VSBinary = ImGUIVS_bytecode_vk;
	auto VSBinarySize = sizeof(ImGUIVS_bytecode_vk);
	auto PSBinary = ImGUIPS_bytecode_vk;
	auto PSBinarySize = sizeof(ImGUIPS_bytecode_vk);
#else 
	auto VSBinary = ImGUIVS_bytecode_dx;
	auto VSBinarySize = sizeof(ImGUIVS_bytecode_dx);
	auto PSBinary = ImGUIPS_bytecode_dx;
	auto PSBinarySize = sizeof(ImGUIPS_bytecode_dx);
#endif 


	// Creating shaders
	m_hVS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_VERTEX), VSBinary, VSBinarySize);
	if (m_hVS == nullptr) return false;

	m_hPS = m_pDevice->createShader(nvrhi::ShaderDesc(nvrhi::ShaderType::SHADER_PIXEL), PSBinary, PSBinarySize);
	if (m_hPS == nullptr) return false;


	// create attribute layout object
	nvrhi::VertexAttributeDesc vertexAttribLayout[] = {
		{ "POSITION", nvrhi::Format::RG32_FLOAT,  1, 0, (size_t)(&((ImDrawVert*)0)->pos), sizeof(ImDrawVert), false },
		{ "TEXCOORD", nvrhi::Format::RG32_FLOAT,  1, 0, (size_t)(&((ImDrawVert*)0)->uv),  sizeof(ImDrawVert), false },
		{ "COLOR", nvrhi::Format::RGBA8_UNORM, 1, 0, (size_t)(&((ImDrawVert*)0)->col), sizeof(ImDrawVert), false },
	};

	m_hAttribLayout = m_pDevice->createInputLayout(vertexAttribLayout, sizeof(vertexAttribLayout) / sizeof(vertexAttribLayout[0]), m_hVS);

	// creating constant buffer
	{
		nvrhi::BufferDesc desc;
		desc.byteSize = sizeof(IMGUI_VS_CBUFFER);
		desc.isConstantBuffer = true;
		desc.debugName = "ImGUI constant buffer";

		m_hCB = m_pDevice->createBuffer(desc);
	}

	// creating font texture
	{
	uint8_t *pixels;
	int width, height;

	io.Fonts->GetTexDataAsRGBA32(&pixels, &width, &height);
		nvrhi::TextureDesc desc;
		desc.width = width;
		desc.height = height;
		desc.format = nvrhi::Format::RGBA8_UNORM;
		desc.debugName = "ImGui font texture";
		m_hFontTexture = m_pDevice->createTexture(desc);
		if (m_hFontTexture == nullptr) return false;

		m_hCommandList->beginTrackingTextureState(m_hFontTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::COMMON);
		m_hCommandList->writeTexture(m_hFontTexture, 0, 0, pixels, width * 4);
		m_hCommandList->endTrackingTextureState(m_hFontTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE, true);

		io.Fonts->TexID = m_hFontTexture;
	}

	// Creating sampler
	{
		nvrhi::SamplerDesc desc;
		desc.wrapMode[0] = nvrhi::SamplerDesc::WRAP_MODE_WRAP;
		desc.wrapMode[1] = nvrhi::SamplerDesc::WRAP_MODE_WRAP;
		desc.wrapMode[2] = nvrhi::SamplerDesc::WRAP_MODE_WRAP;
		desc.mipBias = 0.0;
		desc.anisotropy = 1.0;
		desc.minFilter = true;
		desc.magFilter = true;
		desc.mipFilter = true;

		m_hFontSampler = m_pDevice->createSampler(desc);

		if (m_hFontSampler == nullptr) return false;
	}

	// Creating graphics pipeline
	{
		nvrhi::BlendState blendState;
		blendState.blendEnable[0] = true;
		blendState.srcBlend[0] = nvrhi::BlendState::BLEND_SRC_ALPHA;
		blendState.destBlend[0] = nvrhi::BlendState::BLEND_INV_SRC_ALPHA;
		blendState.blendOp[0] = nvrhi::BlendState::BLEND_OP_ADD;
		blendState.srcBlendAlpha[0] = nvrhi::BlendState::BLEND_INV_SRC_ALPHA;
		blendState.destBlendAlpha[0] = nvrhi::BlendState::BLEND_ZERO;
		blendState.blendOpAlpha[0] = nvrhi::BlendState::BLEND_OP_ADD;
		blendState.colorWriteEnable[0] = nvrhi::BlendState::COLOR_MASK_ALL;

		nvrhi::RasterState rasterState;
		rasterState.fillMode = nvrhi::RasterState::FILL_SOLID;
		rasterState.cullMode = nvrhi::RasterState::CULL_NONE;
		rasterState.scissorEnable = true;
		rasterState.depthClipEnable = true;

		nvrhi::DepthStencilState depthStencilState;
		depthStencilState.depthEnable = false;
		depthStencilState.depthWriteMask = nvrhi::DepthStencilState::DEPTH_WRITE_MASK_ALL;
		depthStencilState.depthFunc = nvrhi::DepthStencilState::COMPARISON_ALWAYS;
		depthStencilState.stencilEnable = false;

		nvrhi::RenderState renderState;
		renderState.blendState = blendState;
		renderState.depthStencilState = depthStencilState;
		renderState.rasterState = rasterState;

		nvrhi::BindingLayoutDesc layoutDesc;
		layoutDesc.VS = { 
			{ 0, nvrhi::ResourceType::ConstantBuffer, 0 } 
		};
		layoutDesc.PS = {
			{ 0, nvrhi::ResourceType::Texture_SRV, 0 },
			{ 0, nvrhi::ResourceType::Sampler, 0 }
		};
		m_hBindingLayout = m_pDevice->createBindingLayout(layoutDesc);

		nvrhi::BindingSetDesc bsDesc;
		bsDesc.VS = { 
			nvrhi::BindingSetItem::ConstantBuffer(0, m_hCB) 
		};
		bsDesc.PS = {
			nvrhi::BindingSetItem::Texture_SRV(0, m_hFontTexture),
			nvrhi::BindingSetItem::Sampler(0, m_hFontSampler)
		};
		m_hBindingSet = m_pDevice->createBindingSet(bsDesc, m_hBindingLayout);



		nvrhi::TextureDesc td;
		td.width = m_backBufferWidth;
		td.height = m_backBufferHeight;
		td.sampleCount = m_sampleCount;
		td.isRenderTarget = true;
		td.clearValue = nvrhi::Color(1.0f);
		td.useClearValue = true;
		td.debugName = "proxy color target";
#if USE_VK
		td.format = nvrhi::Format::BGRA8_UNORM;
#else
		td.format = nvrhi::Format::RGBA8_UNORM;
#endif
		m_hInitTimeRenderTargetTexture = m_pDevice->createTexture(td);
		nvrhi::FramebufferDesc fbDesc;
		fbDesc.addColorAttachment(m_hInitTimeRenderTargetTexture, 0, 0);
		m_hInitTimeFB = m_pDevice->createFramebuffer(fbDesc);

		nvrhi::GraphicsPipelineDesc psoDesc;
		psoDesc.primType = nvrhi::PrimitiveType::TRIANGLE_LIST;
		psoDesc.inputLayout = m_hAttribLayout;
		psoDesc.VS = m_hVS;
		psoDesc.PS = m_hPS;
		psoDesc.renderState = renderState;
		psoDesc.bindingLayouts = { m_hBindingLayout };
		m_hPSO = m_pDevice->createGraphicsPipeline(psoDesc, m_hInitTimeFB);
		assert(m_hPSO);
	}

	io.RenderDrawListsFn = nullptr;
	io.Fonts->AddFontDefault();

	m_hCommandList->close();
	m_pDevice->executeCommandList(m_hCommandList);

	return true;
}

bool UIRenderer::KeyboardUpdate(int key, int scancode, int action, int mods)
{
    auto& io = ImGui::GetIO();

    bool keyIsDown;
    if (action == GLFW_PRESS || action == GLFW_REPEAT)
    {
        keyIsDown = true;
    } else {
        keyIsDown = false;
    }

    // update our internal state tracking for this key button
    m_keyDown[key] = true;

    if (keyIsDown)
    {
        // if the key was pressed, update ImGui immediately
        io.KeysDown[key] = true;
    } else {
        // for key up events, ImGui state is only updated after the next frame
        // this ensures that short keypresses are not missed
    }

    return io.WantCaptureKeyboard;
}

bool UIRenderer::MousePosUpdate(double xpos, double ypos)
{
    auto& io = ImGui::GetIO();
    io.MousePos.x = float(xpos);
    io.MousePos.y = float(ypos);

    return io.WantCaptureMouse;
}

bool UIRenderer::MouseScrollUpdate(double xoffset, double yoffset)
{
    auto& io = ImGui::GetIO();
    io.MouseWheel += float(yoffset);

    return io.WantCaptureMouse;
}

bool UIRenderer::MouseButtonUpdate(int button, int action, int mods)
{
    auto& io = ImGui::GetIO();

    bool buttonIsDown;
    int buttonIndex = 0;

    if (action == GLFW_PRESS || action == GLFW_REPEAT)
    {
        buttonIsDown = true;
    } 
	else 
	{
        buttonIsDown = false;
    }

    switch(button)
    {
        case GLFW_MOUSE_BUTTON_LEFT:
            buttonIndex = 0;
            break;

        case GLFW_MOUSE_BUTTON_RIGHT:
            buttonIndex = 1;
            break;

        case GLFW_MOUSE_BUTTON_MIDDLE:
            buttonIndex = 2;
            break;
    }

    // update our internal state tracking for this mouse button
    m_mouseDown[buttonIndex] = buttonIsDown;

    if (buttonIsDown)
    {
        // update ImGui state immediately
        io.MouseDown[buttonIndex] = true;
    } else {
        // for mouse up events, ImGui state is only updated after the next frame
        // this ensures that short clicks are not missed
    }

    return io.WantCaptureMouse;
}

void UIRenderer::Animate(float elapsedTimeSeconds)
{
    int w, h;
    float scaleX, scaleY;

    m_pDeviceManager->GetWindowDimensions(w, h);
    m_pDeviceManager->GetDPIScaleInfo(scaleX, scaleY);

    ImGuiIO& io = ImGui::GetIO();
    io.DisplaySize = ImVec2(float(w), float(h));
    io.DisplayFramebufferScale.x = scaleX;
    io.DisplayFramebufferScale.y = scaleY;

	io.DeltaTime = elapsedTimeSeconds;
	io.MouseDrawCursor = false;

	ImGui::NewFrame();
}

void UIRenderer::createOrRecreateBufferIfNeeded(nvrhi::BufferHandle& buffer, size_t requiredSize, size_t reallocateSize, const bool indexBuffer)
{
	if (buffer == nullptr || size_t(buffer->GetDesc().byteSize) < requiredSize)
	{
		nvrhi::BufferDesc desc;
		desc.byteSize = uint32_t(reallocateSize);
		desc.structStride = 0;
		desc.debugName = indexBuffer ? "ImGui index buffer" : "ImGui vertex buffer";
		desc.canHaveUAVs = false;
		desc.isVertexBuffer = !indexBuffer;
		desc.isIndexBuffer = indexBuffer;
		desc.isDrawIndirectArgs = false;
		desc.disableGPUsSync = false;
		desc.isVolatile = false;

		buffer = nullptr;
		buffer = m_pDevice->createBuffer(desc);
	}
}

void UIRenderer::updateGeometry()
{
	ImDrawData *drawData = ImGui::GetDrawData();

	// create/resize vertex and index buffers if needed
	createOrRecreateBufferIfNeeded(m_hVB, drawData->TotalVtxCount * sizeof(ImDrawVert), (drawData->TotalVtxCount + 5000) * sizeof(ImDrawVert), false);
	createOrRecreateBufferIfNeeded(m_hIB, drawData->TotalIdxCount * sizeof(ImDrawIdx), (drawData->TotalIdxCount + 5000) * sizeof(ImDrawIdx), true);

	m_IMVertexBuffer.resize(m_hVB->GetDesc().byteSize / sizeof(ImDrawVert));
	m_IMIndexBuffer.resize(m_hIB->GetDesc().byteSize / sizeof(ImDrawIdx));

	// copy and convert all vertices into a single contiguous buffer
	ImDrawVert *vtxDst = m_IMVertexBuffer.data();
	ImDrawIdx *idxDst = m_IMIndexBuffer.data();

	for (int n = 0; n < drawData->CmdListsCount; n++)
	{
		const ImDrawList *cmdList = drawData->CmdLists[n];

		memcpy(vtxDst, cmdList->VtxBuffer.Data, cmdList->VtxBuffer.Size * sizeof(ImDrawVert));
		memcpy(idxDst, cmdList->IdxBuffer.Data, cmdList->IdxBuffer.Size * sizeof(ImDrawIdx));

		vtxDst += cmdList->VtxBuffer.Size;
		idxDst += cmdList->IdxBuffer.Size;
	}
	m_hCommandList->beginTrackingBufferState(m_hVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hIB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->writeBuffer(m_hVB, m_IMVertexBuffer.data(), m_hVB->GetDesc().byteSize);
	m_hCommandList->writeBuffer(m_hIB, m_IMIndexBuffer.data(), m_hIB->GetDesc().byteSize);
	m_hCommandList->endTrackingBufferState(m_hVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->endTrackingBufferState(m_hIB, nvrhi::ResourceStates::COMMON);
}

void UIRenderer::Render(nvrhi::ITexture* renderTarget, uint32_t swapChainIndex)
{
	// Building UI geometry data within ImGUI
    BuildUI();
    ImGui::Render();

	
	ImDrawData *drawData = ImGui::GetDrawData();
	auto& io = ImGui::GetIO();

	m_hCommandList->open();
	m_hCommandList->beginMarker("ImGUI");

	// Getting geometry data from ImGUI
	updateGeometry();


	m_hCommandList->beginTrackingBufferState(m_hCB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingBufferState(m_hIB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->beginTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
	m_hCommandList->beginTrackingTextureState(m_hFontTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);


	// Handling DPI scaling
	drawData->ScaleClipRects(io.DisplayFramebufferScale);

	// Updating constant buffer if needed
	if ((io.DisplaySize.x != m_lastDisplaySizeX) || (io.DisplaySize.y != m_lastDisplaySizeY))
	{

		// setup orthgraphic projection matrix into our constant buffer
		float mvp[4][4] =
		{
			{ 2.0f / io.DisplaySize.x, 0.0f,                   0.0f, 0.0f },
			{ 0.0f,                  2.0f / -io.DisplaySize.y, 0.0f, 0.0f },
			{ 0.0f,                  0.0f,                  -1.0f, 0.0f },
			{ -1.0f,                  1.0f,                   0.0f, 1.0f },
		};

		m_hCommandList->writeBuffer(m_hCB, mvp, sizeof(mvp));
		m_lastDisplaySizeX = (uint32_t)io.DisplaySize.x;
		m_lastDisplaySizeY = (uint32_t)io.DisplaySize.y;
	}


	// Setting up graphics state
	nvrhi::GraphicsState drawState;

	nvrhi::FramebufferHandle hFB = getFramebuffer(renderTarget, swapChainIndex);
	drawState.framebuffer = hFB;
	drawState.pipeline = m_hPSO;
	drawState.bindings = { m_hBindingSet };
	drawState.viewport.viewports.resize(0);
	drawState.viewport.viewports.push_back(nvrhi::Viewport(io.DisplaySize.x * io.DisplayFramebufferScale.x,	io.DisplaySize.y * io.DisplayFramebufferScale.y));
	drawState.viewport.scissorRects.resize(1);  

	nvrhi::VertexBufferBinding vbufBinding;
	vbufBinding.buffer = m_hVB;
	vbufBinding.slot = 0;
	vbufBinding.offset = 0;
	drawState.vertexBuffers.push_back(vbufBinding);

	drawState.indexBuffer.handle = m_hIB;
	drawState.indexBuffer.format = (sizeof(ImDrawIdx) == 2 ? nvrhi::Format::R16_UINT : nvrhi::Format::R32_UINT);
	drawState.indexBuffer.offset = 0;

	// render command lists
	int vtxOffset = 0;
	int idxOffset = 0;
	for (int n = 0; n < drawData->CmdListsCount; n++)
	{
		const ImDrawList *cmdList = drawData->CmdLists[n];
		for (int i = 0; i < cmdList->CmdBuffer.Size; i++)
		{
			const ImDrawCmd *pCmd = &cmdList->CmdBuffer[i];

			if (pCmd->UserCallback)
			{
				pCmd->UserCallback(cmdList, pCmd);
			}
			else {
				
				drawState.viewport.scissorRects[0] = nvrhi::Rect(int(pCmd->ClipRect.x),
					int(pCmd->ClipRect.z),
					int(pCmd->ClipRect.y),
					int(pCmd->ClipRect.w));
					
				nvrhi::DrawArguments drawArguments;
				drawArguments.vertexCount = pCmd->ElemCount;
				drawArguments.startIndexLocation = idxOffset;
				drawArguments.startVertexLocation = vtxOffset;

				m_hCommandList->setGraphicsState(drawState);
				m_hCommandList->drawIndexed(drawArguments);
			}

			idxOffset += pCmd->ElemCount;
		}

		vtxOffset += cmdList->VtxBuffer.Size;
	}

	m_hCommandList->endTrackingBufferState(m_hVB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->endTrackingBufferState(m_hIB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->endTrackingBufferState(m_hCB, nvrhi::ResourceStates::COMMON);
	m_hCommandList->endTrackingTextureState(renderTarget, nvrhi::AllSubresources, nvrhi::ResourceStates::RENDER_TARGET);
	m_hCommandList->endTrackingTextureState(m_hFontTexture, nvrhi::AllSubresources, nvrhi::ResourceStates::SHADER_RESOURCE);

	m_hCommandList->endMarker();
	m_hCommandList->close();
	m_pDevice->executeCommandList(m_hCommandList);


    // reconcile mouse button states
    for(size_t i = 0; i < m_mouseDown.size(); i++)
    {
        if (io.MouseDown[i] == true && m_mouseDown[i] == false)
        {
            io.MouseDown[i] = false;
        }
    }

    // reconcile key states
    for(auto i = 0; i < m_keyDown.size(); i++)
    {
        if (io.KeysDown[i] == true && m_keyDown[i] == false)
        {
            io.KeysDown[i] = false;
        }
    }
}

nvrhi::FramebufferHandle UIRenderer::getFramebuffer(nvrhi::TextureHandle renderTarget, uint32_t swapChainIndex)
{
	nvrhi::TextureDesc td = renderTarget->GetDesc();

	// Create framebuffers if needed
	if (swapChainIndex >= m_hFBs.size()) m_hFBs.resize(swapChainIndex + 1);

	if (m_hFBs[swapChainIndex] == nullptr)
	{
		nvrhi::FramebufferDesc desc = nvrhi::FramebufferDesc().addColorAttachment(renderTarget);

		m_hFBs[swapChainIndex] = m_pDevice->createFramebuffer(desc);
	}

	if (m_needToRecreatePSOs)
	{
		// Recreating PSOs
		nvrhi::GraphicsPipelineDesc PSOdesc = m_hPSO->GetDesc();
		m_hPSO = nullptr;
		m_hPSO = m_pDevice->createGraphicsPipeline(PSOdesc, m_hFBs[swapChainIndex]);
		m_needToRecreatePSOs = false;
	}

	assert(m_hFBs.at(swapChainIndex));
	return m_hFBs[swapChainIndex];
}

void UIRenderer::BackBufferResizing()
{
	m_hFBs.clear();
}
void UIRenderer::BackBufferResized(const uint32_t width, const uint32_t height, const uint32_t sampleCount)
{
	m_backBufferWidth = width;
	m_backBufferHeight = height;
	m_sampleCount = sampleCount;

	if (m_pDevice == nullptr) return;
	if (m_pState == nullptr) return;

	m_pDevice->waitForIdle();

	m_needToRecreatePSOs = true;
}
